/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    Reads .msh format generated by the Adventure system.

    Note: the .msh format does not contain any boundary information. It is
    purely a description of the internal mesh.

    Can read both linear-tet format (i.e. 4 verts per tet) and linear-hex
    format (8 verts per hex) (if provided with the -hex option)
    (Note: will bomb out if not supplied with the correct option for the
     file format)

    Not extensively tested.

\*---------------------------------------------------------------------------*/

#include "argList.H"
#include "Time.H"
#include "polyMesh.H"
#include "IFstream.H"
#include "polyPatch.H"
#include "ListOps.H"
#include "cellModeller.H"

#include <fstream>

using namespace Foam;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

// Main program:

int main(int argc, char *argv[])
{
    argList::noParallel();
    argList::validArgs.append(".msh file");
    argList::validOptions.insert("hex", "");

#   include "setRootCase.H"
#   include "createTime.H"

    bool readHex(args.options().found("hex"));

    fileName mshFile(args.args()[3]);

    IFstream mshStream(mshFile);

    label nCells;

    mshStream >> nCells;

    if (readHex)
    {
        Info<< "Trying to read " << nCells << " hexes." << endl << endl;
    }
    else
    {
        Info<< "Trying to read " << nCells << " tets." << endl << endl;
    }

    cellShapeList cells(nCells);

    const cellModel& tet = *(cellModeller::lookup("tet"));
    const cellModel& hex = *(cellModeller::lookup("hex"));

    labelList tetPoints(4);
    labelList hexPoints(8);

    if (readHex)
    {
        for (label cellI = 0; cellI < nCells; cellI++)
        {
            for (label cp = 0; cp < 8; cp++)
            {
                mshStream >> hexPoints[cp];
            }
            cells[cellI] = cellShape(hex, hexPoints);
        }
    }
    else
    {
        for (label cellI = 0; cellI < nCells; cellI++)
        {
            for (label cp = 0; cp < 4; cp++)
            {
                mshStream >> tetPoints[cp];
            }
            cells[cellI] = cellShape(tet, tetPoints);
        }
    }


    label nPoints;

    mshStream >> nPoints;

    Info<< "Trying to read " << nPoints << " points." << endl << endl;

    pointField points(nPoints);


    for (label pointI = 0; pointI < nPoints; pointI++)
    {
        scalar x, y, z;

        mshStream >> x >> y >> z;

        points[pointI] = point(x, y, z);
    }


    polyMesh mesh
    (
        IOobject
        (
            polyMesh::defaultRegion,
            runTime.constant(),
            runTime
        ),
        points,
        cells,
        faceListList(0),
        wordList(0),
        wordList(0),
        polyPatch::typeName,
        wordList(0)
    );

    Info<< "Writing mesh ..." << endl;

    mesh.write();
    

    Info<< "End\n" << endl;

    return 0;
}


// ************************************************************************* //
